#!/bin/sh
':' //; exec "$(command -v nodejs || command -v node)" "$0" "$@"

/*
	This is a useful script for those managing webide running as a cloud editor,
	it will add users both as system users and into the /etc/webide_users file.
	
	Make this script executable:
	sudo chmod +x adduser.js
	
	Run this script (example: add user foouser with password foopw):
	./adduser.js foouser foopw
	
	Remove a user:
	sudo userdel -r -f nameOfUser
	sudo nano server/users.pw
	
	Gotcha: apparmor can be slooow to update profiles
	
*/

if(process.cwd() != __dirname) throw new Error("adduser.js needs to be run in the same dir as the script eg " + __dirname + " while current directory is " + process.cwd());

console.time("Adding user");

var DEFAULT = require("./server/default_settings.js");

var UTIL = require("./client/UTIL.js");

var fs = require("fs");
var child_process = require('child_process');
var copyFileSync = require("./shared/copyFileSync.js");
var copyFolderRecursiveSync = require("./shared/copyFolderRecursiveSync.js");
var chmodrSync = require("./shared/chmodrSync.js");
var chmodrDirSync = require("./shared/chmodrDirSync.js");
var chownrSync = require("./shared/chownrSync.js");
var chownrDirSync = require("./shared/chownrDirSync.js");

var defaultDomain = DEFAULT.domain;;
var defaultHome = DEFAULT.home_dir;

// Get arguments ...
var getArg = require("./shared/getArg.js");

var username = process.argv[2];
var password = process.argv[3];

if(process.argv[3] == "-c") {
	//console.log(linksTo("/lib/x86_64-linux-gnu/libz.so.1"));
	//process.exit();
	copyNodejs(HOME + username);
	process.exit();
}

var NO_PW_HASH = getArg(["nopwhash"]);
var DOMAIN = getArg(["d", "domain"]) || defaultDomain;
var NOZFS = !!getArg(["nozfs", "nozfs"]);
var HOME = getArg(["home", "home"]) || defaultHome;
var ADMIN_EMAIL = getArg(["email", "email", "mail", "admin_email", "admin_mail"]) || DEFAULT.admin_email;
var NO_CERT = !!getArg(["nocert", "no_cert"]);

if(HOME.charAt(0) != "/") throw new Error("HOME needs to be an absolute path! (start with a slash) HOME=" + HOME);
// Only linux or other unix-like systems are supported (sorry Windows)

if(HOME.charAt(HOME.length-1) != "/") throw new Error("Home dir needs to end with a slash: HOME=" + HOME);

// Favor using JSON as argument to prevent hackers from passing arguments in their password
var maybeJson = process.argv.splice(2, process.argv.length).join(" ");
try { var scriptArguments = JSON.parse(maybeJson); }
catch (err) {
	if(username.charAt(0) == "{") console.log("Unable to parse script parameters as JSON: " + err.message + " maybeJson=" + maybeJson);
var scriptArguments = null; 
}

var UID, GID;

if(scriptArguments) {
	console.log("Using JSON parameters! " + JSON.stringify(scriptArguments, null, 2));
	username = scriptArguments.username;
	password = scriptArguments.password;
	NO_PW_HASH = scriptArguments.noPwHash;
	NO_CERT = scriptArguments.noCert;
	DOMAIN = scriptArguments.domain || defaultDomain;
	HOME = scriptArguments.home || defaultHome;
	UID = scriptArguments.uid || undefined;
	GID = scriptArguments.gid || undefined;
	}


var ENCODING = "utf8";

if(!username) throw new Error("No username specified! scriptArguments=" + scriptArguments + " argv=" + maybeJson);
if(!password) throw new Error("No password specified! scriptArguments=" + scriptArguments + " argv=" + maybeJson);
	
	if(username.match(/[^A-Za-z0-9]/)) throw new Error("Username contains characters that is not a-z or 0-9");

if(username.length < 3) throw new Error("username needs to be at least 3 letters!");
if(username.length > 20) throw new Error("username can not be more then 20 letters!");

console.time("check exist");
// Make sure user not already exist
var userDirs = fs.readdirSync(HOME);
for (var i=0; i<userDirs.length; i++) {
	if(userDirs[i] == username) throw new Error("Directory " + HOME + userDirs[i]+ " already exist!");
}

var etcPasswdString = fs.readFileSync("/etc/passwd", ENCODING);
//console.log("etcPasswdString=" + etcPasswdString);
var users = etcPasswdString.split(/\n|\r\n/);
//console.log("users.length=" + users.length);
var user, name, uid, gid;
var takenUid = [];
var takenGid = [];
for (var i=0; i<users.length; i++) {
	user = users[i].split(":");
	name = user[0];
	uid = parseInt(user[2]);
	gid = parseInt(user[3]);
	if(name == username) throw new Error("User " + username + " already exist in /etc/passwd!");
	takenUid.push(uid);
	takenGid.push(gid);
}

var etcGroupString = fs.readFileSync("/etc/group", ENCODING);
var groups = etcGroupString.split(/\n|\r\n/);
for (var i=0; i<groups.length; i++) {
	group = groups[i].split(":");
	name = group[0];
	gid = parseInt(group[2]);
	if(name == username) throw new Error("Group name " + username + " already exist in /etc/group!");
	takenGid.push(gid);
}

var userHomeDir = HOME + username;
if(fs.existsSync(userHomeDir)) throw new Error("Directory already exist: " + userHomeDir);

console.timeEnd("check exist");



var zfsPool;

if(!NOZFS) {
	console.time("ZFS");
child_process.exec("zfs list", function execAddUser(err, stdout, stderr) {
	// If zfs doesn't exist we get both an err and stderr.
	// If not super user we get both an err and stderr.
	//console.log("zfs: err=" + (err ? err.message : "") + " stdout=" + stdout + " stderr=" + stderr);
	
	if(stderr.indexOf("zfs: not found") != -1) NOZFS = true;
	else if(err) throw err;
	else {
		
			// zpc/home         3.84G  1.37T  5.02M  /home
			var homeWithoutEndingSlashAndEscapedSlashes = HOME.substr(0, HOME.length-1).replace(/\//, "\\/");
			var rePool = new RegExp("(.*)\\/.*" + homeWithoutEndingSlashAndEscapedSlashes + "\\n");
			var matchPool = stdout.match(rePool);
			
	if(matchPool) {
				zfsPool = matchPool[1];
				
				var zfsCreateStdout = child_process.execSync("zfs create " + zfsPool + userHomeDir);
				zfsCreateStdout = zfsCreateStdout.toString(ENCODING);
				
				if(zfsCreateStdout) console.log(zfsCreateStdout);
				else console.log("Created zfs file system on " + userHomeDir);
				
			}
			else {
				console.log("Warning: No zfs file systems exist for " + HOME + " !");
				NOZFS = true;
			}
			}
		
		console.timeEnd("ZFS");
		adduser();
		
});
}
else adduser();

function findFreeUid() {
	var minUid = 1000;
	var maxUid = 60000;
	for (var uid=minUid; uid<maxUid; uid++) {
		if(takenGid.indexOf(uid) != -1) continue; // Try to get the same uid as gid
		if(takenUid.indexOf(uid) == -1) return uid;
	}
	throw new Error("All uid's are taken! Try deleting some users.");
}

function findFreeGid() {
	var minGid = 1000;
	var maxGid = 60000;
	for (var gid=minGid; gid<maxGid; gid++) {
		if(takenGid.indexOf(gid) == -1) return gid;
	}
	throw new Error("All gid's are taken! Try deleting some users/groups.");
}

function createSystemUser(name, uid, gid, homeDir, callback) {
	if(typeof uid == "function" && gid == undefined) {
		callback = uid;
		uid = undefined;
	}
	
	if(callback == undefined) throw new Error("No callback defined!");
	
	if(uid == undefined) uid = findFreeUid();
	if(gid == undefined) gid = findFreeGid();
	
if(homeDir == undefined) {
// Create the home dir
homeDir = HOME + name;
		fs.mkdir(homeDir, function(err) {
			if(err) return callback(err);
			
			addGroupAndUser();
			// No need to chmod or chown as that will be done later in this script
			/*
				fs.chown(homeDir, uid, gid, function(err) {
				if(err) throw err;
				
				fs.chmod(homeDir, mode, function() {
				if(err) throw err;
				addGroupAndUser();
				});
				});
			*/
		});
	}
	else addGroupAndUser();
	
	function addGroupAndUser() {
		
var groupName = name;
		var groupaddCmd = "groupadd -g " + gid + " " + groupName;
		// sudo groupadd -g 144 testgroup
var shell = "/bin/false";
		var useraddCmd = "useradd -d " + homeDir + " -g " + groupName + " -s " + shell + " -u " + uid + " " + name;
		// sudo useradd -d /home/testuser -g testgroup -s /bin/false -u 133 testuser
		
		child_process.exec(groupaddCmd, function execGroupadd(err, stdout, stderr) {
			if(err || stderr) console.log("groupaddCmd=" + groupaddCmd + " err.message=" + (err && err.message) + " stderr=" + stderr);
			
			if (err) return callback(err);
			if(stderr) return callback(new Error(stderr));
			
			console.log("groupadd stdout=" + stdout);
			
			/*
				groupadd: GID '144' already exists
				(Might already have been added to /etc/group)
				
				If groupadd is success it will not return anything
			*/
			
			child_process.exec(useraddCmd, function execUseradd(err, stdout, stderr) {
				if(err || stderr) console.log("useraddCmd=" + useraddCmd + " err.message=" + (err && err.message) + " stderr=" + stderr);
				
				if(stderr) {
					if(stderr.match(/useradd: UID (\d*) is not unique/)) console.warn("Race condition detected!");
					return callback(new Error(stderr));
				}
				
				if (err) return callback(err);
				
				console.log("useradd stdout=" + stdout);
				
				/*
					If useradd is success it will not return anything
				*/
				
				return callback(null, name, uid, gid, homeDir);
				
			});
			
		});
	}
	
	
}


function adduser() {
	
	/*
		We get errors like "useradd: UID 132 is not unique" when using adduser. 
		So we'll use the low level useradd and groupadd instead.
	*/
	
	console.time("create system user");
	createSystemUser(username, UID, GID, NOZFS ? undefined : userHomeDir, function systemUserCreated(err, username, uid, gid, homeDir) {
		if(err) throw err;
		
		console.timeEnd("create system user");
		
		if(typeof uid != "number") throw new Error("uid=" + uid + " needs to be a number. Not a " + (typeof uid));
		if(typeof gid != "number") throw new Error("gid=" + gid + " needs to be a number. Not a " + (typeof gid));
		
		console.log("uid=" + uid);
		console.log("gid=" + gid);
		
		if(homeDir == undefined) throw new Error("createSystemUser called back with homeDir=" + homeDir);
		
		homeDir = UTIL.trailingSlash(homeDir);
		
		//var gid = getGroupId(groupName);
		
		
		// Add skeleton files ...
	copyFolderRecursiveSync("etc/userdir_skeleton/etc", homeDir);
	//copyFolderRecursiveSync("etc/userdir_skeleton/lib", homeDir);
	//copyFolderRecursiveSync("etc/userdir_skeleton/lib64", homeDir);
	copyFolderRecursiveSync("etc/userdir_skeleton/nodejs_examples", homeDir);
	copyFolderRecursiveSync("etc/userdir_skeleton/run", homeDir);
		copyFolderRecursiveSync("etc/userdir_skeleton/ssg_blog_example", homeDir);
	//copyFolderRecursiveSync("etc/userdir_skeleton/usr", homeDir);
		copyFolderRecursiveSync("etc/userdir_skeleton/.webide/", homeDir);
	copyFolderRecursiveSync("etc/userdir_skeleton/wwwpub", homeDir);
		copyFolderRecursiveSync("etc/userdir_skeleton/.ssh/", homeDir);
		
		copyFileSync("etc/userdir_skeleton/.bashrc", homeDir + ".bashrc"); // bash settings, how the prompt look etc
		copyFileSync("etc/userdir_skeleton/.npmrc", homeDir + ".npmrc"); // settings for npm
		
		//copyFileSync("etc/userdir_skeleton/testfile.txt", homeDir + "testfile.txt");
		
	// Use the systems dns settings !?
	//copyFileSync("/run/resolvconf/resolv.conf", homeDir + "run/resolvconf/resolv.conf")
	//copyFileSync("/etc/resolv.conf", homeDir + "etc/resolv.conf")
	
	// Use the systems ca's
	//copyFileSync("/etc/ssl/certs/ca-certifacates.crt", homeDir + "etc/ssl/certs/ca-certifacates.crt")
	
	// The user owns his files
	chownrSync(homeDir, uid, gid);
	
	// Make it so that no one else beside the user can read the user files
	chmodrSync(homeDir, "750");
	
	// home dir needs to have execute permissions for everyone for the unix sockets to work !!!?
	fs.chmodSync(homeDir, "751");
	
	// For DNS lookups to work !?
	chmodrSync(homeDir + "etc/", "444");
	chmodrSync(homeDir + "run/", "444");
	
		// .ssh folder is secret!
		chmodrSync(homeDir + ".ssh/", "700");
		
		
	// Try Copy over the test file (only exist in dev)
	try {
			copyFileSync("./testfile.txt", HOME + username + "/testfile.txt");
	}
	catch(err) {
		if(err.code != "ENOENT") throw err;
	}
	
		// Replace %USERNAME% %HOMEDIR% and %DOMAIN%
		updateFile(homeDir + ".webide/storage/cmsjz_sites");
		updateFile(homeDir + "ssg_blog_example/source/rss_en.xml");
		updateFile(homeDir + "wwwpub/welcome.htm");
		updateFile(homeDir + "nodejs_examples/http_server/http_server_example.js");
		
		
	// add wwwpub
		var wwwgid = getGroupId("www-data");
	//fs.mkdirSync(homeDir + "wwwpub");
		fs.writeFileSync(homeDir + "wwwpub/index.htm", '<!doctype html>\n<meta charset="utf-8">\n\n<body>\n\n<p>Edit me!</p>\n\n</body>\n', ENCODING);
		chownrDirSync(homeDir + "wwwpub", uid, wwwgid);
	// Make wwwpub public, and set the group-id bit so that all new files get the www-data group
		chmodrSync(homeDir + "wwwpub", "2755");
		
		
		// Enable hggit
		fs.writeFileSync(homeDir + ".hgrc", '\n[extensions]\nhgext.bookmarks =\nhggit =\n\n[ui]\nusername = ' + username + "\n\n", ENCODING);
		
		
		// Create a directory for unix sockets
		fs.mkdirSync(homeDir + "sock");
		// Make sure www-data can read and write to unix sockets
		// https://stackoverflow.com/questions/21342828/node-express-unix-domain-socket-permissions
		chmodrSync(homeDir + "sock", "2770"); // Set the group-id bit so that all new files created will belong to the group
		chownrDirSync(homeDir + "sock", uid, wwwgid);
		// note: Each process needs to set umask to give write permission to the group!
		
	// Create a directory where nginx can save logs
	fs.mkdirSync(homeDir + "log");
		chmodrSync(homeDir + "log", "2770"); // Set the group-id bit so that all new files created will belong to the group
		chownrDirSync(homeDir + "log", uid, gid);
		
		// Create a directory for putting "in production" files
		fs.mkdirSync(homeDir + ".prod");
		chmodrSync(homeDir + ".prod", "770");
		chownrDirSync(homeDir + ".prod", uid, gid);
		
		// Create a directory where npm can install packages globally
		fs.mkdirSync(homeDir + ".npm-packages");
		chmodrSync(homeDir + ".npm-packages", "775");
		chownrDirSync(homeDir + ".npm-packages", uid, gid);
	
		
		// Create nginx profile
		var url_user = UTIL.urlFriendly(username);
		
		var nginxProfile = fs.readFileSync("./etc/nginx/user.webide.se.nginx", ENCODING);
		nginxProfile = nginxProfile.replace(/%USERNAME%/g, url_user);
		nginxProfile = nginxProfile.replace(/%HOMEDIR%/g, homeDir);
		nginxProfile = nginxProfile.replace(/%DOMAIN%/g, DOMAIN);
		
	try {
			fs.writeFileSync("/etc/nginx/sites-available/" + url_user + "." + DOMAIN + ".nginx", nginxProfile);
			fs.symlinkSync("/etc/nginx/sites-available/" + url_user + "." + DOMAIN + ".nginx", "/etc/nginx/sites-enabled/" + url_user + "." + DOMAIN + "");
			
			console.time("Reload nginx");
		var reloadNginxStdout = child_process.execSync("service nginx reload");
		reloadNginxStdout = reloadNginxStdout.toString(ENCODING);
			console.timeEnd("Reload nginx");
			if(reloadNginxStdout.trim()) throw new Error(reloadNginxStdout);
	}
	catch (err) {
			console.log(err.message + " Nginx web server is probably not installed. Or there's a problem with the profiles. Try sudo nginx -T && sudo service nginx restart");
			NO_CERT = true;
		}
	
		if(!NO_CERT) {
		// Register SSL certificate for user web page
			console.time("Letsencrypt");
		var letsencrypt = require("./shared/letsencrypt.js");
			letsencrypt.register(url_user + "." + DOMAIN, ADMIN_EMAIL, true);
			console.timeEnd("Letsencrypt");
		}
		
		
		/*
			In order to have a separate apparmor profile for user_worker.js we need a separate executable (scripts also work, but we need the node rcp channel so we need to use fork and not spawn)
			Use mount --bind instead of hard link to prevent EXDEV: cross-device link not permitted (we sometimes got that error even though the link was on the same device!)
			Note: user_worker.js needs capability setgid, setuid, and sys_chroot, which we don't want to give to user scripts, only user_worker.js!
			It's thus important that when user_worker.js forks, it has to set execPath in fork options!!
			
			It will be mounted the first time user logs in.
		*/
		//mount('/usr/bin/node', '/usr/bin/nodejs_' + username);
		
		//var makeNull = child_process.execSync("mknod -m 444 " + HOME + username + "/dev/null c 1 3").toString(ENCODING);
	//if(makeNull.trim() != "") throw makeNull;
		// /dev/null will be created when user first login
		
	// On some systems we need to mount --bind urandom !??
		//mount("/dev/urandom", HOME + username + "/dev/urandom");
		// Will be mounted when the user logs in
	
		
		// These folders need to be created to prevent errors when mounting the binaries (theory why we get errors: creating the folders in parallel leads to racing conditions!?)
		fs.mkdirSync(HOME + username + "/usr/");
		fs.mkdirSync(HOME + username + "/usr/bin/");
	
		fs.chmodSync(HOME + username + "/usr/", "555");
		fs.chmodSync(HOME + username + "/usr/bin/", "555");
		
	
	// npm needs /usr/local/etc or it will try to create it
		fs.mkdirSync(HOME + username + "/usr/local/");
		fs.mkdirSync(HOME + username + "/usr/local/etc");
		
		//chownrDirSync(HOME + username + "/usr/local/", uid, gid);
		//chmodrSync(HOME + username + "/usr/local/", "555");
	

	// Mount these instead of copying to save hdd space
		// They will be mounted when the user first login
		/*
		mount("/lib/", HOME + username + "/lib");
		mount("/lib64/", HOME + username + "/lib64");
		mount("/usr/lib/", HOME + username + "/usr/lib");
		mount("/usr/local/lib", HOME + username + "/usr/local/lib"); // Needed for Python packages (hggit)
		mount("/usr/share/", HOME + username + "/usr/share"); // npm dependencies
		mount("/usr/bin/hg", HOME + username + "/usr/bin/hg");
		mount("/usr/bin/python", HOME + username + "/usr/bin/python");
		mount("/usr/bin/nodejs", HOME + username + "/usr/bin/nodejs");
		mount("/etc/ssl/certs", HOME + username + "/etc/ssl/certs"); // Sometimes? Needed for SSL verfification
		*/
	
		
		// See how to debug apparmor in README.txt
		// Apparmor profiles will be created when the user first login
		/*
		createApparmorProfile("./etc/apparmor/usr.bin.nodejs_someuser", username);
		createApparmorProfile("./etc/apparmor/home.someuser.usr.bin.nodejs", username);
	createApparmorProfile("./etc/apparmor/home.someuser.usr.bin.python", username);
	createApparmorProfile("./etc/apparmor/home.someuser.usr.bin.hg", username);
	createApparmorProfile("./etc/apparmor/home.someuser.usr.share.npm.bin.npm-cli.js", username);
		*/
		
		
		// Activate the user by creating the password file
		if(NO_PW_HASH) {
			var hashedPassword = password;
		}
		else {
			var pwHash = require("./server/pwHash.js");
			var hashedPassword = pwHash(password);
		}
		
		fs.writeFileSync(UTIL.joinPaths([HOME, username, ".webide/", "password"]), hashedPassword, ENCODING);
		
		
		console.log("User with username=" + username + ", password=" + password + ", uid=" + uid + ", gid=" + gid + ", homeDir=" + homeDir + " successfully added! ");
		
		console.timeEnd("Adding user");
		
	//console.log("Wait a few seconds, then sudo service apparmor reload to prevent EACCESS errors");
	
		
		function updateFile(path) {
			var str = fs.readFileSync(path, ENCODING);
			str = str.replace(/%USERNAME%/g, username);
			str = str.replace(/%HOMEDIR%/g, homeDir);
			str = str.replace(/%DOMAIN%/g, DOMAIN);
			fs.writeFileSync(path, str);
		}
		
	});
}

function createApparmorProfile(template, username) {
	/*
		ex: "./etc/apparmor/usr.bin.nodejs_someuser"
	*/
	
	var apparmorProfile = fs.readFileSync(template, ENCODING);
	apparmorProfile = apparmorProfile.replace(/%HOME%/g, HOME);
	apparmorProfile = apparmorProfile.replace(/%USERNAME%/g, username);
	apparmorProfile = apparmorProfile.replace(/%WEBIDE%/g, __dirname);
	
	var dest = template.replace("someuser", username);
	var homeDot = HOME.substr(1).replace(/\//g, "."); // Remove first slash and replace remaining slashes with dots
	dest = dest.replace("home.", homeDot);
	dest = dest.replace("./etc/apparmor/", "/etc/apparmor.d/");
	fs.writeFileSync(dest, apparmorProfile);
	
	var bin = dest.replace("/etc/apparmor.d", "");
	bin = dest.replace(".", "/");
	
	//var enforceApparmorProfileStdout = child_process.execSync("aa-enforce " + bin).toString(ENCODING).trim();
	//if(!enforceApparmorProfileStdout.match(/Setting (.*) to enforce mode./)) throw new Error(enforceApparmorProfileStdout);
}

function mount(source, target) {
	var fs = require("fs");
	
	if ( fs.lstatSync( source ).isDirectory() ) {
		// The source is a directory. Create a directory!
		makeDirPsync(target);
		
	} else {
		// The source is not a directory (it's a file!?). Check if the file exist, then create it
		if ( fs.existsSync( target ) ) throw new Error("File aready exist: " + target); // Prevent overwriting
		fs.closeSync(fs.openSync(target, 'w')); // Create emty file
	}
	
	var mountResult = child_process.execSync("mount --bind " + source + " " + target ).toString(ENCODING).trim();
	if(mountResult != "") throw mountResult;
	
	// Append to /etc/fstab so it is re-mounted after reboot
	//fs.appendFileSync('/etc/fstab', source + '   ' +  target + ' none bind 0 0\n')
	// Server was unable to boot after adding stuff to fstab!!
	
}

function makeDirPsync(target) {
	// Make all the directories in the path ...
	// Should we check if they exist first, or just try creating them all !?
	target = UTIL.trailingSlash(target);
	var paths = UTIL.getFolders(target);
	for (var i=0; i<paths.length; i++) {
		try {
			fs.mkdirSync(paths[i]);
		}
		catch(err) {
			if(err.code != "EEXIST") throw err;
		}
	}
}

function copyProgram(program, homeDir) {
	/*
		Copy a program so it can be spawned from chroot
		Asume the program is located in /usr/bin/
	*/
	
	homeDir = UTIL.trailingSlash(homeDir);
	
	try {
		fs.mkdirSync(homeDir + "usr");
		fs.mkdirSync(homeDir + "usr/bin");
		copyFileSync("/usr/bin/" + program, homeDir + "usr/bin/" + program);
	}
	catch(err) {
		if(err.code != "EEXIST") throw err;
	}
	
	fs.chmodSync(homeDir + "usr/bin/" + program, "555");
	
	// Copy dependencies that the program needs
	var deps = child_process.execSync("ldd " + homeDir + "usr/bin/" + program).toString(ENCODING);
	deps = deps.split(/\n|\r\n/);
	var foldersCreated = [];
	for (var i=0, folders, dir, link; i<deps.length; i++) {
		//console.log(i + ") " + deps[i] + "");
		deps[i] = deps[i].substring(deps[i].indexOf("=>") + 2, deps[i].indexOf("(")-1).trim();
		//console.log(i + ") *" + deps[i] + "*");
		if(deps[i] == "") continue;
		folders = deps[i].split("/");
		dir = homeDir;
		for (var j=1; j<folders.length-1; j++) {
			dir = dir + "/" + folders[j];
			if(foldersCreated.indexOf(dir) != -1) continue;
			try {
				//console.log("Creating folder: " + dir);
				fs.mkdirSync(dir);
				foldersCreated.push(dir);
			}
			catch(err) {
				if(err.code != "EEXIST") throw err;
			}
		}
		// The link to the dep can also be a link itself, so check (recursively) until we find the file
		link = linksTo(deps[i]);
		if(link != null) {
			//console.log("Adding link=" + link);
			deps.push(deps[i] + " => " + link + " ()");
		}
		console.log("Copying " + deps[i]);
		copyFileSync(deps[i], homeDir + deps[i]);
		}
	for (var i=0; i<foldersCreated.length; i++) {
		chmodrSync(foldersCreated[i], "555"); // lib files needs read and execute permission!
	}
}

function linksTo(filePath) {
	try {
		var linkPath = fs.readlinkSync(filePath);
	}
	catch(err) {
		if(err.code == "EINVAL") return null;
		else throw err;
	}
	
	if(linkPath.charAt(0) != "/") {
		// Links to another file in *the same folder*
		var paths = filePath.split("/");
		for (var i=paths.length-2; i>-1; i--) {
			linkPath = paths[i] + "/" + linkPath;
		}
	}
	
	return linkPath;
}

function getGroupId(groupName) {
	var fs = require("fs");
	
	var groupData = fs.readFileSync("/etc/group", ENCODING);
	
	//console.log("groupData=" + groupData);
	
	var groups = groupData.split(/\r|\n/);
	
	// format: groupname:x:115:
	
	for (var i=0, group, name, id; i<groups.length; i++) {
		group = groups[i].split(":");
		name = group[0];
		id = group[2];
		
		if(name == groupName) return parseInt(id);
	}
	
	throw new Error("Unable to find id for groupName=" + groupName);
}

function replaceInFileSync(filePath, arrSearchReplace) {
	var fs = require("fs");
	
	// arrSearchReplace = [searchString, replaceString]
	var text = fs.readFileSync(filePath, ENCODING);
	var maxCount = 500; // Prevent endless loop
	for (var i=0, searchString="", replaceString="", counter=0; i<arrSearchReplace.length; i++) {
		counter = 0;
		while(text.indexOf(searchString) != -1 && counter++ < maxCount) {
		searchString = arrSearchReplace[i][0];
		replaceString = arrSearchReplace[i][1];
		text = text.replace(searchString, replaceString);
			}
	}
	
	fs.writeFileSync(filePath, text, ENCODING);
	
}
